"""
Multi-Layer Perceptron via Sklearn
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This more advanced example shows how sklearn can be used to record an optimization
process in DeepCAVE format.
"""

from sklearn.neural_network import MLPClassifier
from sklearn.model_selection import train_test_split
from ConfigSpace import ConfigurationSpace, Configuration
from ConfigSpace.hyperparameters import (
    UniformFloatHyperparameter,
    CategoricalHyperparameter,
    UniformIntegerHyperparameter,
)
import warnings
from smac import MultiFidelityFacade as MFFacade
from smac import Scenario
from smac.facade import AbstractFacade
from smac.intensifier.hyperband import Hyperband

from sklearn.datasets import load_digits


class MLP:
    def __init__(self):
        self.X_train, self.X_test, self.y_train, self.y_test = self.get_dataset()

    @staticmethod
    def get_dataset():
        digits = load_digits()

        X, y = digits.data, digits.target
        X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=0)

        return X_train, X_test, y_train, y_test

    @staticmethod
    def get_configspace(seed=0):
        configspace = ConfigurationSpace(seed=seed)
        num_neurons_layer1 = UniformIntegerHyperparameter(name="num_neurons_layer1", lower=5, upper=100)
        num_neurons_layer2 = UniformIntegerHyperparameter(name="num_neurons_layer2", lower=5, upper=100)
        activation = CategoricalHyperparameter(name="activation", choices=["logistic", "tanh", "relu"])
        solver = CategoricalHyperparameter(name="solver", choices=["sgd", "adam"])
        batch_size = UniformIntegerHyperparameter(name="batch_size", lower=1, upper=100)
        learning_rate = UniformFloatHyperparameter(
            name="learning_rate", lower=0.0001, upper=0.1, log=True
        )

        configspace.add_hyperparameters(
            [
                num_neurons_layer1,
                num_neurons_layer2,
                activation,
                solver,
                batch_size,
                learning_rate,
            ]
        )

        return configspace

    def train(self, config: Configuration, seed: int = 0, budget: int = 25) -> float:
        # For deactivated parameters (by virtue of the conditions),
        # the configuration stores None-values.
        # This is not accepted by the MLP, so we replace them with placeholder values.

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")

            clf = MLPClassifier(
                max_iter=int(budget),
                hidden_layer_sizes=(
                    config["num_neurons_layer1"],
                    config["num_neurons_layer2"],
                ),
                activation=config["activation"],
                solver=config["solver"],
                batch_size=config["batch_size"],
                learning_rate_init=config["learning_rate"],
            )
            clf.fit(self.X_train, self.y_train)
            score = clf.score(self.X_test, self.y_test)
        return 1 - score


if __name__ == "__main__":
    mlp = MLP()

    facades: list[AbstractFacade] = []
    # Define our environment variables
    scenario = Scenario(
        mlp.get_configspace(),
        walltime_limit=3600,  # After 60 seconds, we stop the hyperparameter optimization
        n_trials=500,  # Evaluate max 500 different trials
        min_budget=25,  # Train the MLP using a hyperparameter configuration for at least 5 epochs
        max_budget=100,  # Train the MLP using a hyperparameter configuration for at most 25 epochs
        n_workers=1,
    )

    print(mlp.train(config=MLP.get_configspace().get_default_configuration()))


    # We want to run five random configurations before starting the optimization.
    initial_design = MFFacade.get_initial_design(scenario, n_configs=5)

    # Create our intensifier
    intensifier = Hyperband(scenario, incumbent_selection="highest_budget", eta=2)

    # Create our SMAC object and pass the scenario and the train method
    smac = MFFacade(
        scenario,
        mlp.train,
        initial_design=initial_design,
        intensifier=intensifier,
        overwrite=True,
        config_selector=MFFacade.get_config_selector(
            scenario=scenario,
            retrain_after= 1
    )
    )

    # Let's optimize
    incumbent = smac.optimize()

    # Get cost of default configuration
    default_cost = smac.validate(mlp.configspace.get_default_configuration())
    print(f"Default cost ({intensifier.__class__.__name__}): {default_cost}")

    # Let's calculate the cost of the incumbent
    incumbent_cost = smac.validate(incumbent)
    print(f"Incumbent cost ({intensifier.__class__.__name__}): {incumbent_cost}")

    facades.append(smac)
